# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/15

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The utilities to cast values into desired types, to avoid IDE warnings.

This module should not import any other module from the application.

"""
from typing import Any


def s(message: Any) -> str:
    """Casts the LazyString message to the string type.

    :param message: The message.
    :return: The binary expression itself.
    """
    return message



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/26

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The queries for the base account management.

"""
import sqlalchemy as sa
from flask import request

from accounting.models import BaseAccount, BaseAccountL10n
from accounting.utils.query import parse_query_keywords


def get_base_account_query() -> list[BaseAccount]:
    """Returns the base accounts, optionally filtered by the query.

    :return: The base accounts.
    """
    keywords: list[str] = parse_query_keywords(request.args.get("q"))
    if len(keywords) == 0:
        return BaseAccount.query.order_by(BaseAccount.code).all()
    conditions: list[sa.BinaryExpression] = []
    for k in keywords:
        l10n: list[BaseAccountL10n] = BaseAccountL10n.query\
            .filter(BaseAccountL10n.title.icontains(k)).all()
        l10n_matches: set[str] = {x.account_code for x in l10n}
        conditions.append(sa.or_(BaseAccount.code.contains(k),
                                 BaseAccount.title_l10n.icontains(k),
                                 BaseAccount.code.in_(l10n_matches)))
    return BaseAccount.query.filter(*conditions)\
        .order_by(BaseAccount.code).all()



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/27

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The common test libraries.

"""
from __future__ import annotations

import re
from typing import Literal

import httpx
from flask import Flask, render_template_string

from accounting.utils.next_uri import encode_next
from test_site import create_app

TEST_SERVER: str = "https://testserver"
"""The test server URI."""
NEXT_URI: str = "/_next"
"""The next URI."""


class Accounts:
    """The shortcuts to the common accounts."""
    CASH: str = "1111-001"
    PETTY_CASH: str = "1112-001"
    BANK: str = "1113-001"
    NOTES_RECEIVABLE: str = "1131-001"
    RECEIVABLE: str = "1141-001"
    MACHINERY: str = "1441-001"
    PREPAID: str = "1258-001"
    NOTES_PAYABLE: str = "2131-001"
    PAYABLE: str = "2141-001"
    SALES: str = "4111-001"
    SERVICE: str = "4611-001"
    AGENCY: str = "4711-001"
    RENT_EXPENSE: str = "6252-001"
    OFFICE: str = "6253-001"
    TRAVEL: str = "6254-001"
    POSTAGE: str = "6256-001"
    UTILITIES: str = "6261-001"
    INSURANCE: str = "6262-001"
    MEAL: str = "6272-001"
    INTEREST: str = "7111-001"
    DONATION: str = "7481-001"
    RENT_INCOME: str = "7482-001"


def create_test_app() -> Flask:
    """Creates and returns the testing Flask application.

    :return: The testing Flask application.
    """
    app: Flask = create_app(is_testing=True)

    @app.get("/.csrf-token")
    def get_csrf_token_view() -> str:
        """The test view to return the CSRF token."""
        return render_template_string("{{csrf_token()}}")

    @app.get("/.messages")
    def get_messages_view() -> str:
        """The test view to return the flashed messages."""
        return render_template_string("{{get_flashed_messages()|tojson}}")

    return app


def get_csrf_token(client: httpx.Client) -> str:
    """Returns the CSRF token.

    :param client: The httpx client.
    :return: The CSRF token.
    """
    return client.get("/.csrf-token").text


def get_client(app: Flask, username: str) -> httpx.Client:
    """Returns a user client.

    :param app: The Flask application.
    :param username: The username.
    :return: The user client.
    """
    client: httpx.Client = httpx.Client(app=app, base_url=TEST_SERVER)
    client.headers["Referer"] = TEST_SERVER
    csrf_token: str = get_csrf_token(client)
    with app.app_context():
        encoded_next_uri: str = encode_next(NEXT_URI)
    response: httpx.Response = client.post("/login",
                                           data={"csrf_token": csrf_token,
                                                 "next": encoded_next_uri,
                                                 "username": username})
    assert response.status_code == 302
    assert response.headers["Location"] == NEXT_URI
    return client


def set_locale(app: Flask, client: httpx.Client, csrf_token: str,
               locale: Literal["en", "zh_Hant", "zh_Hans"]) -> None:
    """Sets the current locale.

    :param app: The Flask application.
    :param client: The test client.
    :param csrf_token: The CSRF token.
    :param locale: The locale.
    :return: None.
    """
    with app.app_context():
        encoded_next_uri: str = encode_next(NEXT_URI)
    response: httpx.Response = client.post("/locale",
                                           data={"csrf_token": csrf_token,
                                                 "locale": locale,
                                                 "next": encoded_next_uri})
    assert response.status_code == 302
    assert response.headers["Location"] == NEXT_URI


def add_journal_entry(client: httpx.Client, form: dict[str, str]) -> int:
    """Adds a transfer journal entry.

    :param client: The client.
    :param form: The form data.
    :return: The newly-added journal entry ID.
    """
    prefix: str = "/accounting/journal-entries"
    journal_entry_type: str = "transfer"
    if len({x for x in form if "-debit-" in x}) == 0:
        journal_entry_type = "receipt"
    elif len({x for x in form if "-credit-" in x}) == 0:
        journal_entry_type = "disbursement"
    store_uri = f"{prefix}/store/{journal_entry_type}"
    response: httpx.Response = client.post(store_uri, data=form)
    assert response.status_code == 302
    return match_journal_entry_detail(response.headers["Location"])


def match_journal_entry_detail(location: str) -> int:
    """Validates if the redirect location is the journal entry detail, and
    returns the journal entry ID on success.

    :param location: The redirect location.
    :return: The journal entry ID.
    :raise AssertionError: When the location is not the journal entry detail.
    """
    m: re.Match = re.match(
        r"^/accounting/journal-entries/(\d+)\?next=", location)
    assert m is not None
    return int(m.group(1))



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/25

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The template filters.

"""
import datetime as dt
from decimal import Decimal
from typing import Any

from flask_babel import get_locale

from accounting.locale import gettext


def format_amount(value: Decimal | None) -> str | None:
    """Formats an amount for readability.

    :param value: The amount.
    :return: The formatted amount text.
    """
    if value is None:
        return None
    if value == 0:
        return "-"
    whole: int = int(value)
    frac: Decimal = (value - whole).normalize()
    return "{:,}".format(whole) + str(abs(frac))[1:]


def format_date(value: dt.date) -> str:
    """Formats a date to be human-friendly.

    :param value: The date.
    :return: The human-friendly date text.
    """
    today: dt.date = dt.date.today()
    if value == today:
        return gettext("Today")
    if value == today - dt.timedelta(days=1):
        return gettext("Yesterday")
    if value == today + dt.timedelta(days=1):
        return gettext("Tomorrow")
    locale = str(get_locale())
    if locale == "zh" or locale.startswith("zh_"):
        if value == today - dt.timedelta(days=2):
            return gettext("The day before yesterday")
        if value == today + dt.timedelta(days=2):
            return gettext("The day after tomorrow")
    if locale == "zh" or locale.startswith("zh_"):
        weekdays = ["一", "二", "三", "四", "五", "六", "日"]
        weekday = weekdays[value.weekday()]
    else:
        weekday = value.strftime("%a")
    if value.year != today.year:
        return "{}/{}/{}({})".format(
            value.year, value.month, value.day, weekday)
    return "{}/{}({})".format(value.month, value.day, weekday)


def default(value: Any, default_value: Any = "") -> Any:
    """Returns the default value if the given value is None.

    :param value: The value.
    :param default_value: The default value when the given value is None.
    :return: The value, or the default value if the given value is None.
    """
    return default_value if value is None else value



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/7

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The utilities to export the report as CSV for download.

"""
import csv
import datetime as dt
from abc import ABC, abstractmethod
from decimal import Decimal
from io import StringIO
from urllib.parse import quote

from flask import Response

from accounting.report.period import Period


class BaseCSVRow(ABC):
    """The base CSV row."""

    @property
    @abstractmethod
    def values(self) -> list[str | Decimal | None]:
        """Returns the values of the row.

        :return: The values of the row.
        """


def csv_download(filename: str, rows: list[BaseCSVRow]) -> Response:
    """Exports the data rows as a CSV file for download.

    :param filename: The download file name.
    :param rows: The data rows.
    :return: The response for download the CSV file.
    """
    with StringIO() as fp:
        writer = csv.writer(fp)
        writer.writerows([x.values for x in rows])
        fp.seek(0)
        response: Response = Response(fp.read(), mimetype="text/csv")
        response.headers["Content-Disposition"] \
            = f"attachment; filename={quote(filename)}"
        return response


def period_spec(period: Period) -> str:
    """Constructs the period specification to be used in the filename.

    :param period: The period.
    :return: The period specification to be used in the filename.
    """
    start: str | None = __get_start_str(period.start)
    end: str | None = __get_end_str(period.end)
    if period.start is None and period.end is None:
        return "all-time"
    if start == end:
        return start
    if period.start is None:
        return f"until-{end}"
    if period.end is None:
        return f"since-{start}"
    return f"{start}-{end}"


def __get_start_str(start: dt.date | None) -> str | None:
    """Returns the string representation of the start date.

    :param start: The start date.
    :return: The string representation of the start date, or None if the start
        date is None.
    """
    if start is None:
        return None
    if start.month == 1 and start.day == 1:
        return str(start.year)
    if start.day == 1:
        return start.strftime("%Y%m")
    return start.strftime("%Y%m%d")


def __get_end_str(end: dt.date | None) -> str | None:
    """Returns the string representation of the end date.

    :param end: The end date.
    :return: The string representation of the end date, or None if the end
        date is None.
    """
    if end is None:
        return None
    if end.month == 12 and end.day == 31:
        return str(end.year)
    if (end + dt.timedelta(days=1)).day == 1:
        return end.strftime("%Y%m")
    return end.strftime("%Y%m%d")



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/25

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The console commands for the base account management.

"""
import csv

import sqlalchemy as sa

from accounting import data_dir
from accounting import db
from accounting.models import BaseAccount, BaseAccountL10n
from accounting.utils.title_case import title_case


def init_base_accounts_command() -> None:
    """Initializes the base accounts."""
    if BaseAccount.query.first() is not None:
        return

    with open(data_dir / "base_accounts.csv") as fp:
        data: list[dict[str, str]] = [x for x in csv.DictReader(fp)]
    account_data: list[dict[str, str]] = [{"code": x["code"],
                                           "title_l10n": title_case(x["title"])}
                                          for x in data]
    locales: list[str] = [x[5:] for x in data[0] if x.startswith("l10n-")]
    l10n_data: list[dict[str, str]] = [{"account_code": x["code"],
                                        "locale": y,
                                        "title": x[f"l10n-{y}"]}
                                       for x in data for y in locales]
    db.session.execute(sa.insert(BaseAccount), account_data)
    db.session.execute(sa.insert(BaseAccountL10n), l10n_data)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/10

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The console commands.

"""
import os

import click
from flask.cli import with_appcontext

from accounting import db
from accounting.account import init_accounts_command
from accounting.base_account import init_base_accounts_command
from accounting.currency import init_currencies_command
from accounting.models import BaseAccount, Account
from accounting.utils.title_case import title_case
from accounting.utils.user import has_user, get_user_pk
import sqlalchemy as sa


def __validate_username(ctx: click.core.Context, param: click.core.Option,
                        value: str) -> str:
    """Validates the username for the click console command.

    :param ctx: The console command context.
    :param param: The console command option.
    :param value: The username.
    :raise click.BadParameter: When validation fails.
    :return: The username.
    """
    value = value.strip()
    if value == "":
        raise click.BadParameter("Username empty.")
    if not has_user(value):
        raise click.BadParameter(f"User {value} does not exist.")
    return value


@click.command("accounting-init-db")
@click.option("-u", "--username", metavar="USERNAME", prompt=True,
              help="The username.", callback=__validate_username,
              default=lambda: os.getlogin())
@with_appcontext
def init_db_command(username: str) -> None:
    """Initializes the accounting database."""
    db.create_all()
    init_base_accounts_command()
    init_accounts_command(username)
    init_currencies_command(username)
    db.session.commit()
    click.echo("Accounting database initialized.")


@click.command("accounting-titleize")
@click.option("-u", "--username", metavar="USERNAME", prompt=True,
              help="The username.", callback=__validate_username,
              default=lambda: os.getlogin())
@with_appcontext
def titleize_command(username: str) -> None:
    """Capitalize the account titles."""
    updater_pk: int = get_user_pk(username)
    updated: int = 0
    for base in BaseAccount.query:
        new_title: str = title_case(base.title_l10n)
        if base.title_l10n != new_title:
            base.title_l10n = new_title
            updated = updated + 1
    for account in Account.query:
        if account.title_l10n.lower() == account.base.title_l10n.lower():
            new_title: str = title_case(account.title_l10n)
            if account.title_l10n != new_title:
                account.title_l10n = new_title
                account.updated_at = sa.func.now()
                account.updated_by_id = updater_pk
                updated = updated + 1
    if updated == 0:
        click.echo("All account titles were already capitalized.")
        return
    db.session.commit()
    click.echo(f"{updated} account titles capitalized.")



# The Mia! Accounting Demonstration Website.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/12

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The data reset for the Mia! Accounting demonstration website.

"""
import datetime as dt

from flask import Flask, Blueprint, url_for, flash, redirect, session, \
    render_template, current_app
from flask_babel import lazy_gettext

from . import db
from .auth import admin_required
from .lib import Accounts, JournalEntryLineItemData, JournalEntryData, \
    JournalEntryCurrencyData, BaseTestData

bp: Blueprint = Blueprint("reset", __name__, url_prefix="/")
"""The blueprint for the data reset."""


@bp.get("reset", endpoint="reset-page")
@admin_required
def reset() -> str:
    """Resets the sample data.

    :return: Redirection to the accounting application.
    """
    return render_template("reset.html")


@bp.post("sample", endpoint="sample")
@admin_required
def reset_sample() -> redirect:
    """Resets the sample data.

    :return: Redirection to the accounting application.
    """
    from accounting.utils.cast import s
    __reset_database()
    SampleData(current_app, "editor").populate()
    flash(s(lazy_gettext(
        "The sample data are emptied and reset successfully.")), "success")
    return redirect(url_for("accounting-report.default"))


@bp.post("reset", endpoint="clean-up")
@admin_required
def clean_up() -> redirect:
    """Clean-up the database data.

    :return: Redirection to the accounting application.
    """
    from accounting.utils.cast import s
    __reset_database()
    flash(s(lazy_gettext("The database is emptied successfully.")), "success")
    return redirect(url_for("accounting-report.default"))


def __reset_database() -> None:
    """Resets the database.

    :return: None.
    """
    from accounting.models import Currency, CurrencyL10n, BaseAccount, \
        BaseAccountL10n, Account, AccountL10n, JournalEntry, \
        JournalEntryLineItem
    from accounting.base_account import init_base_accounts_command
    from accounting.account import init_accounts_command
    from accounting.currency import init_currencies_command

    JournalEntryLineItem.query.delete()
    JournalEntry.query.delete()
    CurrencyL10n.query.delete()
    Currency.query.delete()
    AccountL10n.query.delete()
    Account.query.delete()
    BaseAccountL10n.query.delete()
    BaseAccount.query.delete()
    init_base_accounts_command()
    init_accounts_command(session["user"])
    init_currencies_command(session["user"])
    db.session.commit()


class SampleData(BaseTestData):
    """The sample data."""

    def _init_data(self) -> None:
        self.__add_recurring()
        self.__add_offsets()
        self.__add_meals()

    def __add_recurring(self) -> None:
        """Adds the recurring data.

        :return: None.
        """
        self.__add_usd_recurring()
        self.__add_twd_recurring()

    def __add_usd_recurring(self) -> None:
        """Adds the recurring data in USD.

        :return: None.
        """
        today: dt.date = dt.date.today()
        days: int
        year: int
        month: int

        # Recurring in USD
        date: dt.date = dt.date(today.year - 5, today.month, today.day)
        date = date + dt.timedelta(days=(4 - date.weekday()))
        days = (today - date).days
        while True:
            if days < 0:
                break
            self.__add_journal_entry(
                days, "USD", "2600",
                Accounts.BANK, "Transfer", Accounts.SERVICE, "Payroll")

            days = days - 1
            if days < 0:
                break
            self.__add_journal_entry(
                days, "USD", "1200",
                Accounts.CASH, None, Accounts.BANK, "Withdraw")
            days = days - 13

        year = today.year - 5
        month = today.month
        while True:
            month = month + 1
            if month > 12:
                year = year + 1
                month = 1
            days = (today - dt.date(year, month, 1)).days
            if days < 0:
                break
            self.__add_journal_entry(
                days, "USD", "1800",
                Accounts.RENT_EXPENSE, "Rent", Accounts.BANK, "Transfer")

    def __add_twd_recurring(self) -> None:
        """Adds the recurring data in TWD.

        :return: None.
        """
        today: dt.date = dt.date.today()

        year: int = today.year - 5
        month: int = today.month
        while True:
            days: int = (today - dt.date(year, month, 5)).days
            if days < 0:
                break
            self.__add_journal_entry(
                days, "TWD", "50000",
                Accounts.BANK, "薪資轉帳", Accounts.SERVICE, "薪水")

            days = days - 1
            if days < 0:
                break
            self.__add_journal_entry(
                days, "TWD", "25000",
                Accounts.CASH, None, Accounts.BANK, "提款")

            days = days - 4
            if days < 0:
                break
            self.__add_journal_entry(
                days, "TWD", "18000",
                Accounts.RENT_EXPENSE, "房租", Accounts.BANK, "轉帳")

            month = month + 1
            if month > 12:
                year = year + 1
                month = 1

    def __add_offsets(self) -> None:
        """Adds the offset data.

        :return: None.
        """
        days: int
        year: int
        month: int
        description: str
        line_item_or: JournalEntryLineItemData
        line_item_of: JournalEntryLineItemData

        # Full offset and unmatched in USD
        description = "Speaking—Institute"
        line_item_or = JournalEntryLineItemData(
            Accounts.RECEIVABLE, description, "120")
        self._add_journal_entry(JournalEntryData(
            40, [JournalEntryCurrencyData(
                "USD", [line_item_or], [JournalEntryLineItemData(
                    Accounts.SERVICE, description, "120")])]))
        line_item_of = JournalEntryLineItemData(
            Accounts.RECEIVABLE, description, "120",
            original_line_item=line_item_or)
        self._add_journal_entry(JournalEntryData(
            5, [JournalEntryCurrencyData(
                "USD", [JournalEntryLineItemData(
                    Accounts.BANK, description, "120")],
                [line_item_of])]))
        self.__add_journal_entry(
            30, "USD", "120",
            Accounts.BANK, description, Accounts.SERVICE, description)

        # Partial offset in USD
        line_item_or = JournalEntryLineItemData(
            Accounts.PAYABLE, "Computer", "1600")
        self._add_journal_entry(JournalEntryData(
            60, [JournalEntryCurrencyData(
                "USD", [JournalEntryLineItemData(
                    Accounts.MACHINERY, "Computer", "1600")],
                [line_item_or])]))
        line_item_of = JournalEntryLineItemData(
            Accounts.PAYABLE, "Computer", "800",
            original_line_item=line_item_or)
        self._add_journal_entry(JournalEntryData(
            35, [JournalEntryCurrencyData(
                "USD", [line_item_of], [JournalEntryLineItemData(
                    Accounts.BANK, "Computer", "800")])]))
        line_item_of = JournalEntryLineItemData(
            Accounts.PAYABLE, "Computer", "400",
            original_line_item=line_item_or)
        self._add_journal_entry(JournalEntryData(
            10, [JournalEntryCurrencyData(
                "USD", [line_item_of], [JournalEntryLineItemData(
                    Accounts.CASH, "Computer", "400")])]))

        # Full offset and unmatched in TWD
        description = "演講費—母校"
        line_item_or = JournalEntryLineItemData(
            Accounts.RECEIVABLE, description, "3000")
        self._add_journal_entry(JournalEntryData(
            45, [JournalEntryCurrencyData(
                "TWD", [line_item_or], [JournalEntryLineItemData(
                    Accounts.SERVICE, description, "3000")])]))
        line_item_of = JournalEntryLineItemData(
            Accounts.RECEIVABLE, description, "3000",
            original_line_item=line_item_or)
        self._add_journal_entry(JournalEntryData(
            6, [JournalEntryCurrencyData(
                "TWD", [JournalEntryLineItemData(
                    Accounts.BANK, description, "3000")],
                [line_item_of])]))
        self.__add_journal_entry(
            25, "TWD", "3000",
            Accounts.BANK, description, Accounts.SERVICE, description)

        # Partial offset in TWD
        line_item_or = JournalEntryLineItemData(
            Accounts.PAYABLE, "手機", "30000")
        self._add_journal_entry(JournalEntryData(
            55, [JournalEntryCurrencyData(
                "TWD", [JournalEntryLineItemData(
                    Accounts.MACHINERY, "手機", "30000")],
                [line_item_or])]))
        line_item_of = JournalEntryLineItemData(
            Accounts.PAYABLE, "手機", "16000",
            original_line_item=line_item_or)
        self._add_journal_entry(JournalEntryData(
            27, [JournalEntryCurrencyData(
                "TWD", [line_item_of], [JournalEntryLineItemData(
                    Accounts.BANK, "手機", "16000")])]))
        line_item_of = JournalEntryLineItemData(
            Accounts.PAYABLE, "手機", "6000",
            original_line_item=line_item_or)
        self._add_journal_entry(JournalEntryData(
            8, [JournalEntryCurrencyData(
                "TWD", [line_item_of], [JournalEntryLineItemData(
                    Accounts.CASH, "手機", "6000")])]))

    def __add_meals(self) -> None:
        """Adds the meal data.

        :return: None.
        """
        days = 60
        while days >= 0:
            # Meals in USD
            if days % 4 == 2:
                self.__add_journal_entry(
                    days, "USD", "2.9",
                    Accounts.MEAL, "Lunch—Coffee", Accounts.CASH, None)
            else:
                self.__add_journal_entry(
                    days, "USD", "3.9",
                    Accounts.MEAL, "Lunch—Coffee", Accounts.CASH, None)

            if days % 15 == 3:
                self.__add_journal_entry(
                    days, "USD", "5.45",
                    Accounts.MEAL, "Dinner—Pizza",
                    Accounts.PAYABLE, "Dinner—Pizza")
            else:
                self.__add_journal_entry(
                    days, "USD", "5.9",
                    Accounts.MEAL, "Dinner—Pasta", Accounts.CASH, None)

            # Meals in TWD
            if days % 5 == 3:
                self.__add_journal_entry(
                    days, "TWD", "125",
                    Accounts.MEAL, "午餐—鄰家咖啡", Accounts.CASH, None)
            else:
                self.__add_journal_entry(
                    days, "TWD", "80",
                    Accounts.MEAL, "午餐—便當", Accounts.CASH, None)

            if days % 15 == 3:
                self.__add_journal_entry(
                    days, "TWD", "320",
                    Accounts.MEAL, "晚餐—牛排", Accounts.PAYABLE, "晚餐—牛排")
            else:
                self.__add_journal_entry(
                    days, "TWD", "100",
                    Accounts.MEAL, "晚餐—自助餐", Accounts.CASH, None)

            days = days - 1

    def __add_journal_entry(
            self, days: int, currency: str, amount: str,
            debit_account: str, debit_description: str | None,
            credit_account: str, credit_description: str | None) -> None:
        """Adds a simple journal entry.

        :param days: The number of days before today.
        :param currency: The currency code.
        :param amount: The amount.
        :param debit_account: The debit account code.
        :param debit_description: The debit description.
        :param credit_account: The credit account code.
        :param credit_description: The credit description.
        :return: None.
        """
        self._add_journal_entry(JournalEntryData(
            days,
            [JournalEntryCurrencyData(
                currency,
                [JournalEntryLineItemData(
                    debit_account, debit_description, amount)],
                [JournalEntryLineItemData(
                    credit_account, credit_description, amount)])]))


def init_app(app: Flask) -> None:
    """Initialize the localization.

    :param app: The Flask application.
    :return: None.
    """
    app.register_blueprint(bp)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/25

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The query keyword parser.

This module should not import any other module from the application.

"""
import re


def parse_query_keywords(q: str | None) -> list[str]:
    """Returns the query keywords by the query parameter.

    :param q: The query parameter.
    :return: The query keywords.
    """
    if q is None:
        return []
    q = q.strip()
    if q == "":
        return []
    keywords: list[str] = []
    while True:
        m: re.Match
        m = re.match(r"\"([^\"]+)\"\s+(.+)$", q)
        if m is not None:
            keywords.append(m.group(1))
            q = m.group(2)
            continue
        m = re.match(r"\"([^\"]+)\"?$", q)
        if m is not None:
            keywords.append(m.group(1))
            break
        m = re.match(r"(\S+)\s+(.+)$", q)
        if m is not None:
            keywords.append(m.group(1))
            q = m.group(2)
            continue
        keywords.append(q)
        break
    return keywords



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/30

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The console commands for the account management.

"""
from secrets import randbelow
from typing import Any

import click
import sqlalchemy as sa

from accounting import db
from accounting.models import BaseAccount, Account, AccountL10n
from accounting.utils.user import get_user_pk

AccountData = tuple[int, str, int, str, str, str, bool]
"""The format of the account data, as a list of (ID, base account code, number,
English, Traditional Chinese, Simplified Chinese, is-need-offset) tuples."""


def init_accounts_command(username: str) -> None:
    """Initializes the accounts."""
    creator_pk: int = get_user_pk(username)

    bases: list[BaseAccount] = BaseAccount.query\
        .filter(db.func.length(BaseAccount.code) == 4)\
        .order_by(BaseAccount.code).all()
    if len(bases) == 0:
        raise click.Abort

    existing: list[Account] = Account.query.all()

    existing_base_code: set[str] = {x.base_code for x in existing}
    bases_to_add: list[BaseAccount] = [x for x in bases
                                       if x.code not in existing_base_code]
    if len(bases_to_add) == 0:
        return

    existing_id: set[int] = {x.id for x in existing}

    def get_new_id() -> int:
        """Returns a new random account ID.

        :return: The newly-generated random account ID.
        """
        while True:
            new_id: int = 100000000 + randbelow(900000000)
            if new_id not in existing_id:
                existing_id.add(new_id)
                return new_id

    data: list[dict[str, Any]] = []
    l10n_data: list[dict[str, Any]] = []
    for base in bases_to_add:
        l10n: dict[str, str] = {x.locale: x.title for x in base.l10n}
        account_id: int = get_new_id()
        data.append({"id": account_id,
                     "base_code": base.code,
                     "no": 1,
                     "title_l10n": base.title_l10n,
                     "is_need_offset": __is_need_offset(base.code),
                     "created_by_id": creator_pk,
                     "updated_by_id": creator_pk})
        for locale in {"zh_Hant", "zh_Hans"}:
            l10n_data.append({"account_id": account_id,
                              "locale": locale,
                              "title": l10n[locale]})
    db.session.execute(sa.insert(Account), data)
    db.session.execute(sa.insert(AccountL10n), l10n_data)


def __is_need_offset(base_code: str) -> bool:
    """Checks that whether journal entry line items in the account need offset.

    :param base_code: The code of the base account.
    :return: True if journal entry line items in the account need offset, or
        False otherwise.
    """
    # Assets
    if base_code[0] == "1":
        if base_code[:3] in {"113", "114", "118", "184", "186"}:
            return True
        if base_code in {"1286", "1411", "1421", "1431", "1441", "1511",
                         "1521", "1581", "1611", "1851"}:
            return True
        return False
    # Liabilities
    if base_code[0] == "2":
        if base_code in {"2111", "2114", "2284", "2293", "2861"}:
            return False
        return True
    # Only assets and liabilities need offset
    return False



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/4/8

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The forms for the unmatched offset management.

"""
from decimal import Decimal

import sqlalchemy as sa
from flask_babel import LazyString
from sqlalchemy.orm import selectinload

from accounting.locale import lazy_gettext
from accounting.models import Currency, Account, JournalEntry, \
    JournalEntryLineItem
from accounting.report.utils.unapplied import get_net_balances


class OffsetPair:
    """A pair of an original line item and its offset."""

    def __init__(self, original_line_item: JournalEntryLineItem,
                 offset: JournalEntryLineItem):
        """Constructs a pair of an original line item and its offset.

        :param original_line_item: The original line item.
        :param offset: The offset.
        """
        self.original_line_item: JournalEntryLineItem = original_line_item
        """The original line item."""
        self.offset: JournalEntryLineItem = offset
        """The offset."""


class OffsetMatcher:
    """The offset matcher."""

    def __init__(self, currency: Currency, account: Account):
        """Constructs the offset matcher.

        :param currency: The currency.
        :param account: The account.
        """
        self.__currency: Account = currency
        """The currency."""
        self.__account: Account = account
        """The account."""
        self.matched_pairs: list[OffsetPair] = []
        """A list of matched pairs."""
        self.line_items: list[JournalEntryLineItem] = []
        """The unapplied debits or credits and unmatched offsets."""
        self.unapplied: list[JournalEntryLineItem] = []
        """The unapplied debits or credits."""
        self.unmatched: list[JournalEntryLineItem] = []
        """The unmatched offsets."""
        self.__find_matches()

    def __find_matches(self) -> None:
        """Finds the matched original line items and their offsets.

        :return: None.
        """
        self.__get_line_items()
        if len(self.unapplied) == 0 or len(self.unmatched) == 0:
            return
        remains: list[JournalEntryLineItem] = self.unmatched.copy()
        for original_item in self.unapplied:
            offset_candidates: list[JournalEntryLineItem] \
                = [x for x in remains
                   if (x.journal_entry.date > original_item.journal_entry.date
                       or (x.journal_entry.date
                           == original_item.journal_entry.date
                           and x.journal_entry.no
                           > original_item.journal_entry.no))
                   and x.currency_code == original_item.currency_code
                   and x.description == original_item.description
                   and x.amount == original_item.net_balance]
            if len(offset_candidates) == 0:
                continue
            self.matched_pairs.append(
                OffsetPair(original_item, offset_candidates[0]))
            original_item.match = offset_candidates[0]
            offset_candidates[0].match = original_item
            remains.remove(offset_candidates[0])

    def __get_line_items(self) -> None:
        """Returns the unapplied original line items and unmatched offsets of
        the account.

        :return: The unapplied original line items and unmatched offsets of the
            account.
        """
        net_balances: dict[int, Decimal | None] \
            = get_net_balances(self.__currency, self.__account)
        unmatched_offset_condition: sa.BinaryExpression \
            = sa.and_(Account.id == self.__account.id,
                      JournalEntryLineItem.currency_code
                      == self.__currency.code,
                      JournalEntryLineItem.original_line_item_id.is_(None),
                      sa.or_(sa.and_(Account.base_code.startswith("2"),
                                     JournalEntryLineItem.is_debit),
                             sa.and_(Account.base_code.startswith("1"),
                                     sa.not_(JournalEntryLineItem.is_debit))))
        self.line_items = JournalEntryLineItem.query \
            .join(Account).join(JournalEntry) \
            .filter(sa.or_(JournalEntryLineItem.id.in_(net_balances),
                           unmatched_offset_condition)) \
            .order_by(JournalEntry.date, JournalEntry.no,
                      JournalEntryLineItem.is_debit, JournalEntryLineItem.no) \
            .options(selectinload(JournalEntryLineItem.currency),
                     selectinload(JournalEntryLineItem.journal_entry)).all()
        for line_item in self.line_items:
            line_item.is_offset = line_item.id not in net_balances
        self.unapplied = [x for x in self.line_items if not x.is_offset]
        for line_item in self.unapplied:
            line_item.net_balance = line_item.amount \
                if net_balances[line_item.id] is None \
                else net_balances[line_item.id]
        self.unmatched = [x for x in self.line_items if x.is_offset]
        self.__populate_accumulated_balances()

    def __populate_accumulated_balances(self) -> None:
        """Populates the accumulated balances of the line items.

        :return: None.
        """
        balance: Decimal = Decimal("0")
        for line_item in self.line_items:
            amount: Decimal = line_item.amount if line_item.is_offset \
                else line_item.net_balance
            if line_item.is_debit:
                line_item.debit = amount
                line_item.credit = None
                balance = balance + amount
            else:
                line_item.debit = None
                line_item.credit = amount
                balance = balance - amount
            line_item.balance = balance

    @property
    def status(self) -> str | LazyString:
        """Returns the match status message.

        :return: The match status message.
        """
        if len(self.unmatched) == 0:
            return lazy_gettext("There is no unmatched offset.")
        if len(self.matched_pairs) == 0:
            return lazy_gettext(
                    "%(total)s unmatched offsets without original items.",
                    total=len(self.unmatched))
        return lazy_gettext(
            "%(matches)s unmatched offsets out of %(total)s"
            " can match with their original items.",
            matches=len(self.matched_pairs),
            total=len(self.unmatched))

    def match(self) -> None:
        """Matches the original line items with offsets.

        :return: None.
        """
        for pair in self.matched_pairs:
            pair.offset.original_line_item_id = pair.original_line_item.id



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/6

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The console commands for the currency management.

"""
import csv
from typing import Any

import sqlalchemy as sa

from accounting import db, data_dir
from accounting.models import Currency, CurrencyL10n
from accounting.utils.user import get_user_pk


def init_currencies_command(username: str) -> None:
    """Initializes the currencies."""
    existing_codes: set[str] = {x.code for x in Currency.query.all()}

    with open(data_dir / "currencies.csv") as fp:
        data: list[dict[str, str]] = [x for x in csv.DictReader(fp)]
    to_add: list[dict[str, str]] = [x for x in data
                                    if x["code"] not in existing_codes]
    if len(to_add) == 0:
        return

    creator_pk: int = get_user_pk(username)
    currency_data: list[dict[str, Any]] = [{"code": x["code"],
                                            "name_l10n": x["name"],
                                            "created_by_id": creator_pk,
                                            "updated_by_id": creator_pk}
                                           for x in to_add]
    locales: list[str] = [x[5:] for x in to_add[0] if x.startswith("l10n-")]
    l10n_data: list[dict[str, str]] = [{"currency_code": x["code"],
                                        "locale": y,
                                        "name": x[f"l10n-{y}"]}
                                       for x in to_add for y in locales]
    db.session.execute(sa.insert(Currency), currency_data)
    db.session.execute(sa.insert(CurrencyL10n), l10n_data)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/4

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The report types.

"""
from enum import Enum


class ReportType(Enum):
    """The report types."""
    JOURNAL: str = "journal"
    """The journal."""
    LEDGER: str = "ledger"
    """The ledger."""
    INCOME_EXPENSES: str = "income-expenses"
    """The income and expenses log."""
    TRIAL_BALANCE: str = "trial-balance"
    """The trial balance."""
    INCOME_STATEMENT: str = "income-statement"
    """The income statement."""
    BALANCE_SHEET: str = "balance-sheet"
    """The balance sheet."""
    UNAPPLIED: str = "unapplied"
    """The unapplied original line items."""
    UNMATCHED: str = "unmatched"
    """The unmatched offsets."""
    SEARCH: str = "search"
    """The search."""



# The Mia! Accounting Demonstration Website.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/27

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The Mia! Accounting demonstration website.

"""
import os
from secrets import token_urlsafe
from typing import Type

from click.testing import Result
from flask import Flask, Blueprint, render_template, redirect, Response, \
    url_for
from flask.testing import FlaskCliRunner
from flask_babel_js import BabelJS
from flask_sqlalchemy import SQLAlchemy
from flask_wtf import CSRFProtect
from sqlalchemy import Column

bp: Blueprint = Blueprint("home", __name__)
"""The global blueprint."""
babel_js: BabelJS = BabelJS()
"""The Babel JavaScript instance."""
csrf: CSRFProtect = CSRFProtect()
"""The CSRF protector."""
db: SQLAlchemy = SQLAlchemy()
"""The database instance."""


def create_app(is_testing: bool = False) -> Flask:
    """Create and configure the application.

    :param is_testing: True if we are running for testing, or False otherwise.
    :return: The application.
    """
    import accounting

    app: Flask = Flask(__name__)
    db_uri: str = "sqlite:///" if is_testing else "sqlite:///local.sqlite"
    app.config.from_mapping({
        "SECRET_KEY": os.environ.get("SECRET_KEY", token_urlsafe(32)),
        "SESSION_COOKIE_SAMESITE": "Lax",
        "SESSION_COOKIE_SECURE": True,
        "SQLALCHEMY_DATABASE_URI": db_uri,
        "BABEL_DEFAULT_LOCALE": "en",
        "ALL_LINGUAS": "zh_Hant|正體中文,en|English,zh_Hans|简体中文",
    })
    if is_testing:
        app.config["TESTING"] = True

    babel_js.init_app(app)
    csrf.init_app(app)
    db.init_app(app)

    app.register_blueprint(bp, url_prefix="/")

    from . import locale
    locale.init_app(app)

    from . import auth
    auth.init_app(app)

    from . import reset
    reset.init_app(app)

    class UserUtilities(accounting.UserUtilityInterface[auth.User]):

        def can_view(self) -> bool:
            return auth.current_user() is not None \
                and auth.current_user().username in ["viewer", "editor",
                                                     "admin"]

        def can_edit(self) -> bool:
            return auth.current_user() is not None \
                and auth.current_user().username in ["editor", "admin"]

        def can_admin(self) -> bool:
            return auth.current_user() is not None \
                and auth.current_user().username == "admin"

        def unauthorized(self) -> Response:
            from accounting.utils.next_uri import append_next
            return redirect(append_next(url_for("auth.login-form")))

        @property
        def cls(self) -> Type[auth.User]:
            return auth.User

        @property
        def pk_column(self) -> Column:
            return auth.User.id

        @property
        def current_user(self) -> auth.User | None:
            return auth.current_user()

        def get_by_username(self, username: str) -> auth.User | None:
            return auth.User.query\
                .filter(auth.User.username == username).first()

        def get_pk(self, user: auth.User) -> int:
            return user.id

    accounting.init_app(app, user_utils=UserUtilities())

    with app.app_context():
        init_db(app)

    return app


def init_db(app: Flask) -> None:
    """Initializes the database.

    :param app: The Flask application.
    :return: None.
    """
    db.create_all()
    from .auth import User
    for username in ["viewer", "editor", "admin", "nobody"]:
        if User.query.filter(User.username == username).first() is None:
            db.session.add(User(username=username))
    db.session.commit()
    runner: FlaskCliRunner = app.test_cli_runner()
    result: Result = runner.invoke(args=["accounting-init-db", "-u", "editor"])
    assert result.exit_code == 0, result.output + str(result.exception)


@bp.get("/", endpoint="home")
def get_home() -> str:
    """Returns the home page.

    :return: The home page.
    """
    return render_template("home.html")



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/4

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The utility to return the end of a month.

"""
import calendar
import datetime as dt


def month_end(day: dt.date) -> dt.date:
    """Returns the end day of month for a date.

    :param day: The date.
    :return: The end day of the month of that day.
    """
    last_day: int = calendar.monthrange(day.year, day.month)[1]
    return dt.date(day.year, day.month, last_day)



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/30

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The views for the account management.

"""
from urllib.parse import parse_qsl, urlencode

import sqlalchemy as sa
from flask import Blueprint, render_template, session, redirect, flash, \
    url_for, request
from werkzeug.datastructures import ImmutableMultiDict

from accounting import db
from accounting.locale import lazy_gettext
from accounting.models import Account, BaseAccount
from accounting.utils.cast import s
from accounting.utils.flash_errors import flash_form_errors
from accounting.utils.next_uri import inherit_next, or_next
from accounting.utils.pagination import Pagination
from accounting.utils.permission import can_view, has_permission, can_edit
from accounting.utils.user import get_current_user_pk
from .forms import AccountForm, sort_accounts_in, AccountReorderForm
from .queries import get_account_query

bp: Blueprint = Blueprint("account", __name__)
"""The view blueprint for the account management."""


@bp.get("", endpoint="list")
@has_permission(can_view)
def list_accounts() -> str:
    """Lists the accounts.

    :return: The account list.
    """
    accounts: list[BaseAccount] = get_account_query()
    pagination: Pagination = Pagination[BaseAccount](accounts)
    return render_template("accounting/account/list.html",
                           list=pagination.list, pagination=pagination)


@bp.get("create", endpoint="create")
@has_permission(can_edit)
def show_add_account_form() -> str:
    """Shows the form to add an account.

    :return: The form to add an account.
    """
    if "form" in session:
        form = AccountForm(ImmutableMultiDict(parse_qsl(session["form"])))
        del session["form"]
        form.validate()
    else:
        form = AccountForm()
    return render_template("accounting/account/create.html",
                           form=form)


@bp.post("store", endpoint="store")
@has_permission(can_edit)
def add_account() -> redirect:
    """Adds an account.

    :return: The redirection to the account detail on success, or the account
        creation form on error.
    """
    form = AccountForm(request.form)
    if not form.validate():
        flash_form_errors(form)
        session["form"] = urlencode(list(request.form.items()))
        return redirect(inherit_next(url_for("accounting.account.create")))
    account: Account = Account()
    form.populate_obj(account)
    db.session.add(account)
    db.session.commit()
    flash(s(lazy_gettext("The account is added successfully")), "success")
    return redirect(inherit_next(__get_detail_uri(account)))


@bp.get("<account:account>", endpoint="detail")
@has_permission(can_view)
def show_account_detail(account: Account) -> str:
    """Shows the account detail.

    :param account: The account.
    :return: The detail.
    """
    return render_template("accounting/account/detail.html", obj=account)


@bp.get("<account:account>/edit", endpoint="edit")
@has_permission(can_edit)
def show_account_edit_form(account: Account) -> str:
    """Shows the form to edit an account.

    :param account: The account.
    :return: The form to edit the account.
    """
    form: AccountForm
    if "form" in session:
        form = AccountForm(ImmutableMultiDict(parse_qsl(session["form"])))
        del session["form"]
        form.validate()
    else:
        form = AccountForm(obj=account)
    return render_template("accounting/account/edit.html",
                           account=account, form=form)


@bp.post("<account:account>/update", endpoint="update")
@has_permission(can_edit)
def update_account(account: Account) -> redirect:
    """Updates an account.

    :param account: The account.
    :return: The redirection to the account detail on success, or the account
        edit form on error.
    """
    form = AccountForm(request.form)
    if not form.validate():
        flash_form_errors(form)
        session["form"] = urlencode(list(request.form.items()))
        return redirect(inherit_next(url_for("accounting.account.edit",
                                             account=account)))
    with db.session.no_autoflush:
        form.populate_obj(account)
    if not account.is_modified:
        flash(s(lazy_gettext("The account was not modified.")), "success")
        return redirect(inherit_next(__get_detail_uri(account)))
    account.updated_by_id = get_current_user_pk()
    account.updated_at = sa.func.now()
    db.session.commit()
    flash(s(lazy_gettext("The account is updated successfully.")), "success")
    return redirect(inherit_next(__get_detail_uri(account)))


@bp.post("<account:account>/delete", endpoint="delete")
@has_permission(can_edit)
def delete_account(account: Account) -> redirect:
    """Deletes an account.

    :param account: The account.
    :return: The redirection to the account list on success, or the account
        detail on error.
    """
    if not account.can_delete:
        flash(s(lazy_gettext("The account cannot be deleted.")), "error")
        return redirect(inherit_next(__get_detail_uri(account)))
    account.delete()
    sort_accounts_in(account.base_code, account.id)
    db.session.commit()
    flash(s(lazy_gettext("The account is deleted successfully.")), "success")
    return redirect(or_next(__get_list_uri()))


@bp.get("bases/<baseAccount:base>", endpoint="order")
@has_permission(can_view)
def show_account_order(base: BaseAccount) -> str:
    """Shows the order of the accounts under a same base account.

    :param base: The base account.
    :return: The order of the accounts under the base account.
    """
    return render_template("accounting/account/order.html", base=base)


@bp.post("bases/<baseAccount:base>", endpoint="sort")
@has_permission(can_edit)
def sort_accounts(base: BaseAccount) -> redirect:
    """Reorders the accounts under a base account.

    :param base: The base account.
    :return: The redirection to the incoming account or the account list.  The
        reordering operation does not fail.
    """
    form: AccountReorderForm = AccountReorderForm(base)
    form.save_order()
    if not form.is_modified:
        flash(s(lazy_gettext("The order was not modified.")), "success")
        return redirect(or_next(__get_list_uri()))
    db.session.commit()
    flash(s(lazy_gettext("The order is updated successfully.")), "success")
    return redirect(or_next(__get_list_uri()))


def __get_detail_uri(account: Account) -> str:
    """Returns the detail URI of an account.

    :param account: The account.
    :return: The detail URI of the account.
    """
    return url_for("accounting.account.detail", account=account)


def __get_list_uri() -> str:
    """Returns the account list URI.

    :return: The account list URI.
    """
    return url_for("accounting.account.list")



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/3/22

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The forms for the option management.

"""
from flask import render_template
from flask_babel import LazyString
from flask_wtf import FlaskForm
from wtforms import StringField, FieldList, FormField, IntegerField
from wtforms.validators import DataRequired, ValidationError

from accounting.forms import ACCOUNT_REQUIRED, CurrencyExists, AccountExists, \
    IsDebitAccount, IsCreditAccount
from accounting.locale import lazy_gettext
from accounting.models import Account
from accounting.utils.current_account import CurrentAccount
from accounting.utils.options import Options
from accounting.utils.strip_text import strip_text


class CurrentAccountExists:
    """The validator to check that the current account exists."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        if field.data is None or field.data == CurrentAccount.CURRENT_AL_CODE:
            return
        if Account.find_by_code(field.data) is None:
            raise ValidationError(lazy_gettext(
                "The account does not exist."))


class AccountNotCurrent:
    """The validator to check that the account is a current account."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        if field.data is None or field.data == CurrentAccount.CURRENT_AL_CODE:
            return
        if field.data[:2] not in {"11", "12", "21", "22"}:
            raise ValidationError(lazy_gettext(
                "This is not a current account."))


class NotStartPayableFromExpense:
    """The validator to check that a payable line item does not start from
    expense."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        if field.data is None or field.data[0] != "2":
            return
        account: Account | None = Account.find_by_code(field.data)
        if account is not None and account.is_need_offset:
            raise ValidationError(lazy_gettext(
                "You cannot select a payable account as expense."))


class NotStartReceivableFromIncome:
    """The validator to check that a receivable line item does not start
    from income."""

    def __call__(self, form: FlaskForm, field: StringField) -> None:
        if field.data is None or field.data[0] != "1":
            return
        account: Account | None = Account.find_by_code(field.data)
        if account is not None and account.is_need_offset:
            raise ValidationError(lazy_gettext(
                "You cannot select a receivable account as income."))


class RecurringItemForm(FlaskForm):
    """The base sub-form to add or update the recurring item."""
    no = IntegerField()
    """The order number of this recurring item."""
    name = StringField()
    """The name of the recurring item."""
    account_code = StringField()
    """The account code."""
    description_template = StringField()
    """The description template."""

    @property
    def account_text(self) -> str | None:
        """Returns the account text.

        :return: The account text.
        """
        if self.account_code.data is None:
            return None
        account: Account | None = Account.find_by_code(self.account_code.data)
        return None if account is None else str(account)

    @property
    def all_errors(self) -> list[str | LazyString]:
        """Returns all the errors of the form.

        :return: All the errors of the form.
        """
        all_errors: list[str | LazyString] = []
        for key in self.errors:
            if key != "csrf_token":
                all_errors.extend(self.errors[key])
        return all_errors


class RecurringExpenseForm(RecurringItemForm):
    """The sub-form to add or update the recurring expenses."""
    no = IntegerField()
    """The order number of this recurring item."""
    name = StringField(
        filters=[strip_text],
        validators=[DataRequired(lazy_gettext("Please fill in the name."))])
    """The name of the recurring item."""
    account_code = StringField(
        filters=[strip_text],
        validators=[
            ACCOUNT_REQUIRED,
            AccountExists(),
            IsDebitAccount(lazy_gettext("This account is not for expense.")),
            NotStartPayableFromExpense()])
    """The account code."""
    description_template = StringField(
        filters=[strip_text],
        validators=[
            DataRequired(lazy_gettext(
                "Please fill in the description template."))])
    """The template for the line item description."""


class RecurringIncomeForm(RecurringItemForm):
    """The sub-form to add or update the recurring incomes."""
    no = IntegerField()
    """The order number of this recurring item."""
    name = StringField(
        filters=[strip_text],
        validators=[DataRequired(lazy_gettext("Please fill in the name."))])
    """The name of the recurring item."""
    account_code = StringField(
        filters=[strip_text],
        validators=[
            ACCOUNT_REQUIRED,
            AccountExists(),
            IsCreditAccount(lazy_gettext("This account is not for income.")),
            NotStartReceivableFromIncome()])
    """The account code."""
    description_template = StringField(
        filters=[strip_text],
        validators=[
            DataRequired(lazy_gettext(
                "Please fill in the description template."))])
    """The description template."""


class RecurringForm(RecurringItemForm):
    """The sub-form for the recurring expenses and incomes."""
    expenses = FieldList(FormField(RecurringExpenseForm), name="expense")
    """The recurring expenses."""
    incomes = FieldList(FormField(RecurringIncomeForm), name="income")
    """The recurring incomes."""

    @property
    def item_template(self) -> str:
        """Returns the template of a recurring item.

        :return: The template of a recurring item.
        """
        return render_template(
            "accounting/option/include/form-recurring-item.html",
            expense_income="EXPENSE_INCOME",
            item_index="ITEM_INDEX",
            form=RecurringItemForm())

    @property
    def expense_accounts(self) -> list[Account]:
        """The expense accounts.

        :return: None.
        """
        return Account.selectable_debit()

    @property
    def income_accounts(self) -> list[Account]:
        """The income accounts.

        :return: None.
        """
        return Account.selectable_credit()

    @property
    def as_data(self) -> dict[str, list[tuple[str, str, str]]]:
        """Returns the form data.

        :return: The form data.
        """
        def as_tuple(item: RecurringItemForm) -> tuple[str, str, str]:
            return (item.name.data, item.account_code.data,
                    item.description_template.data)

        expenses: list[RecurringItemForm] = [x.form for x in self.expenses]
        self.__sort_item_forms(expenses)
        incomes: list[RecurringItemForm] = [x.form for x in self.incomes]
        self.__sort_item_forms(incomes)
        return {"expense": [as_tuple(x) for x in expenses],
                "income": [as_tuple(x) for x in incomes]}

    @staticmethod
    def __sort_item_forms(forms: list[RecurringItemForm]) -> None:
        """Sorts the recurring item sub-forms.

        :param forms: The recurring item sub-forms.
        :return: None.
        """
        ord_by_form: dict[RecurringItemForm, int] \
            = {forms[i]: i for i in range(len(forms))}
        recv_no: set[int] = {x.no.data for x in forms if x.no.data is not None}
        missing_recv_no: int = 100 if len(recv_no) == 0 else max(recv_no) + 100
        forms.sort(key=lambda x: (x.no.data or missing_recv_no,
                                  ord_by_form.get(x)))


class OptionForm(FlaskForm):
    """The form to update the options."""
    default_currency_code = StringField(
        filters=[strip_text],
        validators=[
            DataRequired(lazy_gettext("Please select the default currency.")),
            CurrencyExists()])
    """The default currency code."""
    default_ie_account_code = StringField(
        filters=[strip_text],
        validators=[
            DataRequired(lazy_gettext(
                "Please select the default account"
                " for the income and expenses log.")),
            CurrentAccountExists(),
            AccountNotCurrent()])
    """The default account code for the income and expenses log."""
    recurring = FormField(RecurringForm)
    """The recurring expenses and incomes."""

    def populate_obj(self, obj: Options) -> None:
        """Populates the form data into a currency object.

        :param obj: The currency object.
        :return: None.
        """
        obj.default_currency_code = self.default_currency_code.data
        obj.default_ie_account_code = self.default_ie_account_code.data
        obj.recurring_data = self.recurring.form.as_data

    @property
    def current_accounts(self) -> list[CurrentAccount]:
        """Returns the current accounts.

        :return: The current accounts.
        """
        return CurrentAccount.accounts()



# The Mia! Accounting Project.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/2/28

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The test for the description editor.

"""
import datetime as dt
import unittest

import httpx
from flask import Flask

from accounting.utils.next_uri import encode_next
from testlib import NEXT_URI, Accounts, create_test_app, get_client, \
    get_csrf_token, add_journal_entry


class DescriptionEditorTestCase(unittest.TestCase):
    """The description editor test case."""

    def setUp(self) -> None:
        """Sets up the test.
        This is run once per test.

        :return: None.
        """
        self.__app: Flask = create_test_app()
        """The Flask application."""

        with self.__app.app_context():
            from accounting.models import JournalEntry, JournalEntryLineItem
            JournalEntry.query.delete()
            JournalEntryLineItem.query.delete()
            self.__encoded_next_uri: str = encode_next(NEXT_URI)
            """The encoded next URI."""

        self.__client: httpx.Client = get_client(self.__app, "editor")
        """The user client."""
        self.__csrf_token: str = get_csrf_token(self.__client)
        """The CSRF token."""

    def test_description_editor(self) -> None:
        """Test the description editor.

        :return: None.
        """
        from accounting.journal_entry.utils.description_editor import \
            DescriptionEditor
        for form in get_form_data(self.__csrf_token, self.__encoded_next_uri):
            add_journal_entry(self.__client, form)
        with self.__app.app_context():
            editor: DescriptionEditor = DescriptionEditor()

        # Debit-General
        self.assertEqual(len(editor.debit.general.tags), 2)
        self.assertEqual(editor.debit.general.tags[0].name, "Lunch")
        self.assertEqual(len(editor.debit.general.tags[0].accounts), 2)
        self.assertEqual(editor.debit.general.tags[0].accounts[0].code,
                         Accounts.MEAL)
        self.assertEqual(editor.debit.general.tags[0].accounts[1].code,
                         Accounts.PETTY_CASH)
        self.assertEqual(editor.debit.general.tags[1].name, "Dinner")
        self.assertEqual(len(editor.debit.general.tags[1].accounts), 2)
        self.assertEqual(editor.debit.general.tags[1].accounts[0].code,
                         Accounts.MEAL)
        self.assertEqual(editor.debit.general.tags[1].accounts[1].code,
                         Accounts.PETTY_CASH)

        # Debit-Travel
        self.assertEqual(len(editor.debit.travel.tags), 3)
        self.assertEqual(editor.debit.travel.tags[0].name, "Bike")
        self.assertEqual(len(editor.debit.travel.tags[0].accounts), 1)
        self.assertEqual(editor.debit.travel.tags[0].accounts[0].code,
                         Accounts.TRAVEL)
        self.assertEqual(editor.debit.travel.tags[1].name, "Taxi")
        self.assertEqual(len(editor.debit.travel.tags[1].accounts), 1)
        self.assertEqual(editor.debit.travel.tags[1].accounts[0].code,
                         Accounts.TRAVEL)
        self.assertEqual(editor.debit.travel.tags[2].name, "Airplane")
        self.assertEqual(len(editor.debit.travel.tags[2].accounts), 1)
        self.assertEqual(editor.debit.travel.tags[2].accounts[0].code,
                         Accounts.TRAVEL)

        # Debit-Bus
        self.assertEqual(len(editor.debit.bus.tags), 2)
        self.assertEqual(editor.debit.bus.tags[0].name, "Train")
        self.assertEqual(len(editor.debit.bus.tags[0].accounts), 1)
        self.assertEqual(editor.debit.bus.tags[0].accounts[0].code,
                         Accounts.TRAVEL)
        self.assertEqual(editor.debit.bus.tags[1].name, "Bus")
        self.assertEqual(len(editor.debit.bus.tags[1].accounts), 1)
        self.assertEqual(editor.debit.bus.tags[1].accounts[0].code,
                         Accounts.TRAVEL)

        # Credit-General
        self.assertEqual(len(editor.credit.general.tags), 2)
        self.assertEqual(editor.credit.general.tags[0].name, "Lunch")
        self.assertEqual(len(editor.credit.general.tags[0].accounts), 3)
        self.assertEqual(editor.credit.general.tags[0].accounts[0].code,
                         Accounts.PETTY_CASH)
        self.assertEqual(editor.credit.general.tags[0].accounts[1].code,
                         Accounts.BANK)
        self.assertEqual(editor.credit.general.tags[0].accounts[2].code,
                         Accounts.CASH)
        self.assertEqual(editor.credit.general.tags[1].name, "Dinner")
        self.assertEqual(len(editor.credit.general.tags[1].accounts), 2)
        self.assertEqual(editor.credit.general.tags[1].accounts[0].code,
                         Accounts.BANK)
        self.assertEqual(editor.credit.general.tags[1].accounts[1].code,
                         Accounts.PETTY_CASH)

        # Credit-Travel
        self.assertEqual(len(editor.credit.travel.tags), 2)
        self.assertEqual(editor.credit.travel.tags[0].name, "Bike")
        self.assertEqual(len(editor.credit.travel.tags[0].accounts), 2)
        self.assertEqual(editor.credit.travel.tags[0].accounts[0].code,
                         Accounts.PETTY_CASH)
        self.assertEqual(editor.credit.travel.tags[0].accounts[1].code,
                         Accounts.PREPAID)
        self.assertEqual(editor.credit.travel.tags[1].name, "Taxi")
        self.assertEqual(len(editor.credit.travel.tags[1].accounts), 2)
        self.assertEqual(editor.credit.travel.tags[1].accounts[0].code,
                         Accounts.PETTY_CASH)
        self.assertEqual(editor.credit.travel.tags[1].accounts[1].code,
                         Accounts.CASH)

        # Credit-Bus
        self.assertEqual(len(editor.credit.bus.tags), 2)
        self.assertEqual(editor.credit.bus.tags[0].name, "Train")
        self.assertEqual(len(editor.credit.bus.tags[0].accounts), 2)
        self.assertEqual(editor.credit.bus.tags[0].accounts[0].code,
                         Accounts.PREPAID)
        self.assertEqual(editor.credit.bus.tags[0].accounts[1].code,
                         Accounts.PETTY_CASH)
        self.assertEqual(editor.credit.bus.tags[1].name, "Bus")
        self.assertEqual(len(editor.credit.bus.tags[1].accounts), 1)
        self.assertEqual(editor.credit.bus.tags[1].accounts[0].code,
                         Accounts.PREPAID)


def get_form_data(csrf_token: str, encoded_next_uri: str) \
        -> list[dict[str, str]]:
    """Returns the form data for multiple journal entry forms.

    :param csrf_token: The CSRF token.
    :param encoded_next_uri: The encoded next URI.
    :return: A list of the form data.
    """
    journal_entry_date: str = dt.date.today().isoformat()
    return [{"csrf_token": csrf_token,
             "next": encoded_next_uri,
             "date": journal_entry_date,
             "currency-0-code": "USD",
             "currency-0-credit-0-account_code": Accounts.SERVICE,
             "currency-0-credit-0-description": " Salary ",
             "currency-0-credit-0-amount": "2500"},
            {"csrf_token": csrf_token,
             "next": encoded_next_uri,
             "date": journal_entry_date,
             "currency-0-code": "USD",
             "currency-0-debit-0-account_code": Accounts.MEAL,
             "currency-0-debit-0-description": " Lunch—Fish ",
             "currency-0-debit-0-amount": "4.7",
             "currency-0-credit-0-account_code": Accounts.BANK,
             "currency-0-credit-0-description": " Lunch—Fish ",
             "currency-0-credit-0-amount": "4.7",
             "currency-0-debit-1-account_code": Accounts.MEAL,
             "currency-0-debit-1-description": " Lunch—Fries ",
             "currency-0-debit-1-amount": "2.15",
             "currency-0-credit-1-account_code": Accounts.PETTY_CASH,
             "currency-0-credit-1-description": " Lunch—Fries ",
             "currency-0-credit-1-amount": "2.15",
             "currency-0-debit-2-account_code": Accounts.MEAL,
             "currency-0-debit-2-description": " Dinner—Hamburger ",
             "currency-0-debit-2-amount": "4.25",
             "currency-0-credit-2-account_code": Accounts.BANK,
             "currency-0-credit-2-description": " Dinner—Hamburger ",
             "currency-0-credit-2-amount": "4.25"},
            {"csrf_token": csrf_token,
             "next": encoded_next_uri,
             "date": journal_entry_date,
             "currency-0-code": "USD",
             "currency-0-debit-0-account_code": Accounts.MEAL,
             "currency-0-debit-0-description": " Lunch—Salad ",
             "currency-0-debit-0-amount": "4.99",
             "currency-0-credit-0-account_code": Accounts.CASH,
             "currency-0-credit-0-description": " Lunch—Salad ",
             "currency-0-credit-0-amount": "4.99",
             "currency-0-debit-1-account_code": Accounts.MEAL,
             "currency-0-debit-1-description": " Dinner—Steak  ",
             "currency-0-debit-1-amount": "8.28",
             "currency-0-credit-1-account_code": Accounts.PETTY_CASH,
             "currency-0-credit-1-description": " Dinner—Steak ",
             "currency-0-credit-1-amount": "8.28"},
            {"csrf_token": csrf_token,
             "next": encoded_next_uri,
             "date": journal_entry_date,
             "currency-0-code": "USD",
             "currency-0-debit-0-account_code": Accounts.MEAL,
             "currency-0-debit-0-description": " Lunch—Pizza  ",
             "currency-0-debit-0-amount": "5.49",
             "currency-0-credit-0-account_code": Accounts.PETTY_CASH,
             "currency-0-credit-0-description": " Lunch—Pizza ",
             "currency-0-credit-0-amount": "5.49",
             "currency-0-debit-1-account_code": Accounts.MEAL,
             "currency-0-debit-1-description": " Lunch—Noodles ",
             "currency-0-debit-1-amount": "7.47",
             "currency-0-credit-1-account_code": Accounts.PETTY_CASH,
             "currency-0-credit-1-description": " Lunch—Noodles ",
             "currency-0-credit-1-amount": "7.47"},
            {"csrf_token": csrf_token,
             "next": encoded_next_uri,
             "date": journal_entry_date,
             "currency-0-code": "USD",
             "currency-0-debit-0-account_code": Accounts.TRAVEL,
             "currency-0-debit-0-description": " Airplane—Lake City↔Hill Town",
             "currency-0-debit-0-amount": "800"},
            {"csrf_token": csrf_token,
             "next": encoded_next_uri,
             "date": journal_entry_date,
             "currency-0-code": "USD",
             "currency-0-debit-0-account_code": Accounts.TRAVEL,
             "currency-0-debit-0-description": " Bus—323—Downtown→Museum ",
             "currency-0-debit-0-amount": "2.5",
             "currency-0-credit-0-account_code": Accounts.PREPAID,
             "currency-0-credit-0-description": " Bus—323—Downtown→Museum ",
             "currency-0-credit-0-amount": "2.5",
             "currency-0-debit-1-account_code": Accounts.TRAVEL,
             "currency-0-debit-1-description": " Train—Blue—Museum→Central ",
             "currency-0-debit-1-amount": "3.2",
             "currency-0-credit-1-account_code": Accounts.PREPAID,
             "currency-0-credit-1-description": " Train—Blue—Museum→Central ",
             "currency-0-credit-1-amount": "3.2",
             "currency-0-debit-2-account_code": Accounts.TRAVEL,
             "currency-0-debit-2-description": " Train—Green—Central→Mall ",
             "currency-0-debit-2-amount": "3.2",
             "currency-0-credit-2-account_code": Accounts.PREPAID,
             "currency-0-credit-2-description": " Train—Green—Central→Mall ",
             "currency-0-credit-2-amount": "3.2",
             "currency-0-debit-3-account_code": Accounts.TRAVEL,
             "currency-0-debit-3-description": " Train—Red—Mall→Museum ",
             "currency-0-debit-3-amount": "4.4",
             "currency-0-credit-3-account_code": Accounts.PETTY_CASH,
             "currency-0-credit-3-description": " Train—Red—Mall→Museum ",
             "currency-0-credit-3-amount": "4.4"},
            {"csrf_token": csrf_token,
             "next": encoded_next_uri,
             "date": journal_entry_date,
             "currency-0-code": "USD",
             "currency-0-debit-0-account_code": Accounts.TRAVEL,
             "currency-0-debit-0-description": " Taxi—Museum→Office ",
             "currency-0-debit-0-amount": "15.5",
             "currency-0-credit-0-account_code": Accounts.CASH,
             "currency-0-credit-0-description": " Taxi—Museum→Office ",
             "currency-0-credit-0-amount": "15.5",
             "currency-0-debit-1-account_code": Accounts.TRAVEL,
             "currency-0-debit-1-description": " Taxi—Office→Restaurant ",
             "currency-0-debit-1-amount": "12",
             "currency-0-credit-1-account_code": Accounts.PETTY_CASH,
             "currency-0-credit-1-description": " Taxi—Office→Restaurant ",
             "currency-0-credit-1-amount": "12",
             "currency-0-debit-2-account_code": Accounts.TRAVEL,
             "currency-0-debit-2-description": " Taxi—Restaurant→City Hall ",
             "currency-0-debit-2-amount": "8",
             "currency-0-credit-2-account_code": Accounts.PETTY_CASH,
             "currency-0-credit-2-description": " Taxi—Restaurant→City Hall ",
             "currency-0-credit-2-amount": "8",
             "currency-0-debit-3-account_code": Accounts.TRAVEL,
             "currency-0-debit-3-description": " Bike—City Hall→Office ",
             "currency-0-debit-3-amount": "3.5",
             "currency-0-credit-3-account_code": Accounts.PETTY_CASH,
             "currency-0-credit-3-description": " Bike—City Hall→Office ",
             "currency-0-credit-3-amount": "3.5",
             "currency-0-debit-4-account_code": Accounts.TRAVEL,
             "currency-0-debit-4-description": " Bike—Restaurant→Office ",
             "currency-0-debit-4-amount": "4",
             "currency-0-credit-4-account_code": Accounts.PETTY_CASH,
             "currency-0-credit-4-description": " Bike—Restaurant→Office ",
             "currency-0-credit-4-amount": "4",
             "currency-0-debit-5-account_code": Accounts.TRAVEL,
             "currency-0-debit-5-description": " Bike—Office→Theatre ",
             "currency-0-debit-5-amount": "1.5",
             "currency-0-credit-5-account_code": Accounts.PETTY_CASH,
             "currency-0-credit-5-description": " Bike—Office→Theatre ",
             "currency-0-credit-5-amount": "1.5",
             "currency-0-debit-6-account_code": Accounts.TRAVEL,
             "currency-0-debit-6-description": " Bike—Theatre→Home ",
             "currency-0-debit-6-amount": "5.5",
             "currency-0-credit-6-account_code": Accounts.PREPAID,
             "currency-0-credit-6-description": " Bike—Theatre→Home ",
             "currency-0-credit-6-amount": "5.5"},
            {"csrf_token": csrf_token,
             "next": encoded_next_uri,
             "date": journal_entry_date,
             "currency-0-code": "USD",
             "currency-0-debit-0-account_code": Accounts.PETTY_CASH,
             "currency-0-debit-0-description": " Dinner—Steak  ",
             "currency-0-debit-0-amount": "8.28",
             "currency-0-credit-0-account_code": Accounts.BANK,
             "currency-0-credit-0-description": " Dinner—Steak ",
             "currency-0-credit-0-amount": "8.28",
             "currency-0-debit-1-account_code": Accounts.PETTY_CASH,
             "currency-0-debit-1-description": " Lunch—Pizza ",
             "currency-0-debit-1-amount": "5.49",
             "currency-0-credit-1-account_code": Accounts.BANK,
             "currency-0-credit-1-description": " Lunch—Pizza  ",
             "currency-0-credit-1-amount": "5.49"}]



# The Mia! Accounting Demonstration Website.
# Author: imacat@mail.imacat.idv.tw (imacat), 2023/1/2

#  Copyright (c) 2023 imacat.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""The localization for the Mia! Accounting demonstration website.

"""
from babel import Locale
from flask import request, session, current_app, Blueprint, Response, \
    redirect, Flask
from flask_babel import Babel
from werkzeug.datastructures import LanguageAccept

from accounting.utils.next_uri import or_next

bp: Blueprint = Blueprint("locale", __name__, url_prefix="/")
"""The blueprint for the localization."""


def get_locale():
    """Returns the locale of the user

    :return: The locale of the user.
    """
    all_linguas: dict[str, str] = get_all_linguas()
    if "locale" in session and session["locale"] in all_linguas:
        return session["locale"]
    return __fix_accept_language(request.accept_languages)\
        .best_match(all_linguas.keys())


def __fix_accept_language(accept: LanguageAccept) -> LanguageAccept:
    """Fixes the accept-language so that territory variants may be matched to
    script variants.  For example, zh_TW, zh_HK to zh_Hant, and zh_CN, zh_SG to
    zh_Hans.  This is to solve the issue that Flask only recognizes the script
    variants, like zh_Hant and zh_Hans.

    :param accept: The original HTTP accept languages.
    :return: The fixed HTTP accept languages
    """
    accept_list: list[tuple[str, float]] = list(accept)
    to_add: list[tuple[str, float]] = []
    for pair in accept_list:
        locale: Locale = Locale.parse(pair[0].replace("-", "_"))
        if locale.script is not None:
            tag: str = f"{locale.language}-{locale.script}"
            if tag not in accept:
                to_add.append((tag, pair[1]))
    accept_list.extend(to_add)
    return LanguageAccept(accept_list)


@bp.post("/locale", endpoint="set-locale")
def set_locale() -> Response:
    """Sets the locale for the user.

    :return: The response.
    """
    all_linguas: dict[str, str] = get_all_linguas()
    if "locale" in request.form and request.form["locale"] in all_linguas:
        session["locale"] = request.form["locale"]
    return redirect(or_next("/"))


def get_all_linguas() -> dict[str, str]:
    """Returns all the available languages.

    :return: All the available languages, as a dictionary of the language code
        and their local names.
    """
    return {y[0]: y[1] for y in
            [x.split("|") for x in
             current_app.config["ALL_LINGUAS"].split(",")]}


def init_app(app: Flask) -> None:
    """Initialize the localization.

    :param app: The Flask application.
    :return: None.
    """
    babel = Babel()
    babel.init_app(app, locale_selector=get_locale)
    app.register_blueprint(bp)
    app.jinja_env.globals["get_locale"] = get_locale
    app.jinja_env.globals["get_all_linguas"] = get_all_linguas
